use crate::i18ndictionary::I18NDictionary;

pub struct Names {
    family_name: I18NDictionary,
    designer: I18NDictionary,
    designer_url: I18NDictionary,
    manufacturer: I18NDictionary,
    manufacturer_url: I18NDictionary,
    license: I18NDictionary,
    license_url: I18NDictionary,
    version: I18NDictionary,
    unique_id: I18NDictionary,
    description: I18NDictionary,
    typographic_family: I18NDictionary,
    typographic_subfamily: I18NDictionary,
    compatible_full_name: I18NDictionary,
    sample_text: I18NDictionary,
    w_w_s_family_name: I18NDictionary,
    w_w_s_subfamily_name: I18NDictionary,
    copyright: I18NDictionary,
    style_map_family_name: I18NDictionary,
    trademark: I18NDictionary,
}
